%/<copyright-disclaimer-start>
%/  **************************************************************************************************************
%/  * © 2025 Advanced Micro Devices, Inc. All rights reserved.                                                   *
%/  * DISCLAIMER                                                                                                 *
%/  * The information contained herein is for informational purposes only, and is subject to change              *
%/  * without notice. While every precaution has been taken in the preparation of this document, it              *
%/  * may contain technical inaccuracies, omissions and typographical errors, and AMD is under no                *
%/  * obligation to update or otherwise correct this information.  Advanced Micro Devices, Inc. makes            *
%/  * no representations or warranties with respect to the accuracy or completeness of the contents of           *
%/  * this document, and assumes no liability of any kind, including the implied warranties of noninfringement,  *
%/  * merchantability or fitness for particular purposes, with respect to the operation or use of AMD            *
%/  * hardware, software or other products described herein.  No license, including implied or                   *
%/  * arising by estoppel, to any intellectual property rights is granted by this document.  Terms and           *
%/  * limitations applicable to the purchase or use of AMD’s products are as set forth in a signed agreement     *
%/  * between the parties or in AMD's Standard Terms and Conditions of Sale. GD-18                               *
%/  *                                                                                                            *
%/  **************************************************************************************************************
%/<copyright-disclaimer-end>


clear all;
close all;
rng(1);

%% Generate input stimuli
TT = numerictype(1,16,15);
FF = fimath('RoundingMethod','Round','OverflowAction','Saturate');
fs = 100;          % Sampling frequency in MHz
bw_sig0 = 0.35;    % Signal Bandwidth, normalized to fs
bw_sig1 = 0.25;    % Signal Bandwidth, normalized to fs
delta_sig = 0.05;  % 1/2 Signal transition band, normalized to fs

Nsamp = 256;
fir_taps0 = firgr('minevn',[0,bw_sig0-delta_sig,bw_sig0+delta_sig,1],[1,1,0,0],[1e-4,1e-4]);
backoff_dB = -20;                       % To achieve (-1,+1) range on AWGN signal
fir_i0 = sqrt(0.5)*10^(0.05*backoff_dB)*complex(randn(1,Nsamp),randn(1,Nsamp));
fir_o0 = filter(fir_taps0,1,fir_i0);
sig_i0_fxp = fi(fir_o0,TT,FF);
sig_i0 = double(sig_i0_fxp);  % Quantize input signal

fir_taps1 = firgr('minevn',[0,bw_sig1-delta_sig,bw_sig1+delta_sig,1],[1,1,0,0],[1e-4,1e-4]);
backoff_dB = -20;                       % To achieve (-1,+1) range on AWGN signal
fir_i1 = sqrt(0.5)*10^(0.05*backoff_dB)*complex(randn(1,Nsamp),randn(1,Nsamp));
fir_o1 = filter(fir_taps1,1,fir_i1);
sig_i1_fxp = fi(fir_o1,TT,FF);
sig_i1 = double(sig_i1_fxp);  % Quantize input signal

% Save the signals for AIE simulation (assuming plio_64_bits):
[~,~,~] = mkdir('data');
fid = fopen('data/input_sig_i0.txt','w');
for ii = 1 : 2 :Nsamp
  fprintf(fid,'%d %d %d %d\n',...
          real(sig_i0_fxp(ii+0)).int,imag(sig_i0_fxp(ii+0)).int,...
          real(sig_i0_fxp(ii+1)).int,imag(sig_i0_fxp(ii+1)).int);
end
fclose(fid);

fid = fopen('data/input_sig_i1.txt','w');
for ii = 1 : 2 : Nsamp
  fprintf(fid,'%d %d %d %d\n',...
          real(sig_i1_fxp(ii+0)).int,imag(sig_i1_fxp(ii+0)).int,...
          real(sig_i1_fxp(ii+1)).int,imag(sig_i1_fxp(ii+1)).int);
end
fclose(fid);

%% Fractional Rate Conversion (3/2)

% Fp  = fs/2 + 0.1*fs;     % passband-edge frequency
% Fst = fs   - 0.1*fs;     % stopband-edge frequency
Fp  = 0.8*fs;              % passband-edge frequency
Fst = 2*fs   - Fp;         % stopband-edge frequency
Rp  = 1e-4;                % Ripple
Rst = 1e-4;                % Stopband attenuation
frc_taps_ori = 3*firgr('minevn',[0,Fp,Fst,3*fs]/(3*fs),[1,1,0,0],[Rp,Rst]);
frc_taps_fxp = fi(frc_taps_ori,TT,FF);
frc_taps = double(frc_taps_fxp);
range = linspace(-0.5*3*fs,0.5*3*fs,1e3);
[H,f] = freqz(frc_taps,1,range,3*fs);
figure(1);
subplot(2,1,1);plot(f,20*log10(abs(H)));grid on;xlabel('Frequency (MHz)');ylabel('Magnitude (dB)');title(sprintf('FRC 3/2 # taps = %d, Magnitude',length(frc_taps)));
subplot(2,1,2);plot(f,unwrap(angle(H)));grid on;xlabel('Frequency (MHz)');ylabel('Phase (degrees)');title(sprintf('FRC 3/2 # taps = %d, Phase',length(frc_taps)));

% Store resampler taps for AIE implementation:
fid = fopen('data/resamp_coeff.h','w');
fprintf(fid,'#define RESAMP_COEFF {\\\n');
last = (length(frc_taps_ori));
for tt = 1 : last
  if ( tt == last ) fprintf(fid,'%d }\n',frc_taps_fxp.int(tt));
  else              fprintf(fid,'%d,\\\n',frc_taps_fxp.int(tt));
  end
end
fclose(fid);


%% HBF1
Fp  = 0.8*fs;              % passband-edge frequency
Fst = 2*3/2*fs - Fp;       % stopband-edge frequency
Rp  = 1e-4;                % Ripple
Rst = 1e-4;                % Stopband attenuation
hb1_taps_ori = 2*firgr('minevn',[0,Fp,Fst,2*3/2*fs]/(2*3/2*fs),[1,1,0,0],[Rp,Rst]);
hb1_taps_fxp = fi(hb1_taps_ori,TT,FF);  %fxp
hb1_taps = double(hb1_taps_fxp);    %flt
range = linspace(-0.5*3/2*2*fs,0.5*3/2*2*fs,1e3);
[H,f] = freqz(hb1_taps,1,range,3/2*2*fs);
figure(2);
subplot(2,1,1);plot(f,20*log10(abs(H)));grid on;xlabel('Frequency (MHz)');ylabel('Magnitude (dB)');title(sprintf('HBF1 # taps = %d, Magnitude',length(hb1_taps)));
subplot(2,1,2);plot(f,unwrap(angle(H)));grid on;xlabel('Frequency (MHz)');ylabel('Phase (degrees)');title(sprintf('HBF1 # taps = %d, Phase',length(hb1_taps)));

% Store HB1 taps for AIE implementation:
fid = fopen('data/hb1_coeff.h','w');
fprintf(fid,'#define HB1_COEFF {\\\n');
last = (length(hb1_taps_ori));
for tt = 1 : last
  if ( tt == last ) fprintf(fid,'%d }\n',hb1_taps_fxp.int(tt));
  else              fprintf(fid,'%d,\\\n',hb1_taps_fxp.int(tt));
  end
end
fclose(fid);


%% HBF2
Fp  = 0.8*fs;              % passband-edge frequency
Fst = 2*2*3/2*fs - Fp;     % stopband-edge frequency
Rp  = 1e-4;                % Ripple
Rst = 1e-4;                % Stopband attenuation
hb2_taps_ori = 2*firgr('minevn',[0,Fp,Fst,2*2*3/2*fs]/(2*2*3/2*fs),[1,1,0,0],[Rp,Rst]);
hb2_taps_fxp = fi(hb2_taps_ori,TT,FF);
hb2_taps = double(hb2_taps_fxp);
range = linspace(-0.5*3/2*2*2*fs,0.5*3/2*2*2*fs,1e3);
[H,f] = freqz(hb2_taps,1,range,3/2*2*2*fs);
figure(3);
subplot(2,1,1);plot(f,20*log10(abs(H)));grid on;xlabel('Frequency (MHz)');ylabel('Magnitude (dB)');title(sprintf('HBF2 # taps = %d, Magnitude',length(hb2_taps)));
subplot(2,1,2);plot(f,unwrap(angle(H)));grid on;xlabel('Frequency (MHz)');ylabel('Phase (degrees)');title(sprintf('HBF2 # taps = %d, Phase',length(hb2_taps)));

% Store HB2 taps for AIE implementation:
fid = fopen('data/hb2_coeff.h','w');
fprintf(fid,'#define HB2_COEFF {\\\n');
last = (length(hb2_taps_ori));
for tt = 1 : last
  if ( tt == last ) fprintf(fid,'%d }\n',hb2_taps_fxp.int(tt));
  else              fprintf(fid,'%d,\\\n',hb2_taps_fxp.int(tt));
  end
end
fclose(fid);


%% HBF3
Fp  = 0.8*fs;              % passband-edge frequency
Fst = 2*2*2*3/2*fs - Fp;   % stopband-edge frequency
Rp  = 1e-4;                % Ripple
Rst = 1e-4;                % Stopband attenuation
hb3_taps_ori = 2*firgr('minevn',[0,Fp,Fst,2*2*2*3/2*fs]/(2*2*2*3/2*fs),[1,1,0,0],[Rp,Rst]);
hb3_taps_fxp = fi(hb3_taps_ori,TT,FF);
hb3_taps = double(hb3_taps_fxp);
range = linspace(-0.5*3/2*2*2*2*fs,0.5*3/2*2*2*2*fs,1e3);
[H,f] = freqz(hb3_taps,1,range,3/2*2*2*2*fs);
figure(4);
subplot(2,1,1);plot(f,20*log10(abs(H)));grid on;xlabel('Frequency (MHz)');ylabel('Magnitude (dB)');title(sprintf('HBF3 # taps = %d, Magnitude',length(hb3_taps)));
subplot(2,1,2);plot(f,unwrap(angle(H)));grid on;xlabel('Frequency (MHz)');ylabel('Phase (degrees)');title(sprintf('HBF3 # taps = %d, Phase',length(hb3_taps)));

% Store HB3 taps for AIE implementation:
fid = fopen('data/hb3_coeff.h','w');
fprintf(fid,'#define HB3_COEFF {\\\n');
last = (length(hb3_taps_ori));
for tt = 1 : last
  if ( tt == last ) fprintf(fid,'%d }\n',hb3_taps_fxp.int(tt));
  else              fprintf(fid,'%d,\\\n',hb3_taps_fxp.int(tt));
  end
end
fclose(fid);

%% Mixer
fc0 = 200;                  % Tuner frequency in MHz
fc1 = -350;                 % Tuner frequency in MHz

%% MATLAB model

frc_o0_fxp = fi(downsample(filter(frc_taps,1,double(upsample(sig_i0,3))),2),TT,FF);      % RCF 3/2
frc_o0 = double(frc_o0_fxp);                                                             % RCF 3/2
hb1_o0_fxp = fi(filter(hb1_taps,1,upsample(frc_o0,2)),TT,FF);                            % ↑2
hb1_o0 = double(hb1_o0_fxp);                                                             % ↑2
hb2_o0_fxp = fi(filter(hb2_taps,1,upsample(hb1_o0,2)),TT,FF);                            % ↑2
hb2_o0 = double(hb2_o0_fxp);                                                             % ↑2
hb3_o0_fxp = fi(filter(hb3_taps,1,upsample(hb2_o0,2)),TT,FF);  
hb3_o0 = double(hb3_o0_fxp);                                                           % ↑2
tuner_o0_fxp = fi(hb3_o0.*exp(1i*2*pi*fc0/(fs*3/2*2*2*2)*[0:length(hb3_o0)-1]),TT,FF);   % Tuner
tuner_o0 = double(tuner_o0_fxp);   % Tuner

frc_o1_fxp = fi(downsample(filter(frc_taps,1,double(upsample(sig_i1,3))),2),TT,FF);      % RCF 3/2
frc_o1 = double(frc_o1_fxp);                                                             % RCF 3/2
hb1_o1_fxp = fi(filter(hb1_taps,1,upsample(frc_o1,2)),TT,FF);                            % ↑2
hb1_o1 = double(hb1_o1_fxp);                                                             % ↑2
hb2_o1_fxp = fi(filter(hb2_taps,1,upsample(hb1_o1,2)),TT,FF);                            % ↑2
hb2_o1 = double(hb2_o1_fxp);                                                             % ↑2
hb3_o1_fxp = fi(filter(hb3_taps,1,upsample(hb2_o1,2)),TT,FF);                            % ↑2
hb3_o1 = double(hb3_o1_fxp);                                                             % ↑2
tuner_o1_fxp = fi(hb3_o1.*exp(1i*2*pi*fc1/(fs*3/2*2*2*2)*[0:length(hb3_o1)-1]),TT,FF);   % Tuner
tuner_o1 = double(tuner_o1_fxp);   % Tuner

sig_o_fxp = fi(tuner_o0+tuner_o1,TT,FF);                                             % Sum and quantize output
sig_o = double(sig_o_fxp);                                                           % Sum and quantize output

% test 
mixer0_phaeinc = fi(exp(1i*2*pi*fc1/(fs*3/2*2*2*2)*[0:length(hb3_o0)-1]));
mixer1_phaeinc = exp(1i*2*pi*fc1/(fs*3/2*2*2*2)*[0:length(hb3_o1)-1]);

% %%%%

% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Save the ouput signals for sig 0
% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Resampler output
fid = fopen('data/output_resamp_0.txt','w');
% Two samples in one line
for ii = 1 : 2 :Nsamp*(3/2)
  fprintf(fid,'%d %d %d %d\n',...
          real(frc_o0_fxp(ii+0)).int,imag(frc_o0_fxp(ii+0)).int,...
          real(frc_o0_fxp(ii+1)).int,imag(frc_o0_fxp(ii+1)).int);
end
fclose(fid);

% One sample in one line
fid = fopen('data/output_resamp_1_one.txt','w');
for ii = 1 : Nsamp*(3/2)
  fprintf(fid,'%d %d \n',...
          real(frc_o0_fxp(ii+0)).int,imag(frc_o0_fxp(ii+0)).int);
end
fclose(fid);

% HB1 output
fid = fopen('data/output_hb1_0.txt','w');
for ii = 1 : 2 : Nsamp*(3/2)*2
  fprintf(fid,'%d %d %d %d\n',...
          real(hb1_o0_fxp(ii+0)).int,imag(hb1_o0_fxp(ii+0)).int,...;
          real(hb1_o0_fxp(ii+1)).int,imag(hb1_o0_fxp(ii+1)).int);
end
fclose(fid);

% HB2 output
fid = fopen('data/output_hb2_0.txt','w');
for ii = 1 : 2 : Nsamp*(3/2)*2*2
  fprintf(fid,'%d %d %d %d\n',...
          real(hb2_o0_fxp(ii+0)).int,imag(hb2_o0_fxp(ii+0)).int,...
          real(hb2_o0_fxp(ii+1)).int,imag(hb2_o0_fxp(ii+1)).int);
end
fclose(fid);

% HB3 output
fid = fopen('data/output_hb3_0.txt','w');
for ii = 1 : 2 : Nsamp*(3/2)*2*2*2
  fprintf(fid,'%d %d %d %d\n',...
          real(hb3_o0_fxp(ii+0)).int,imag(hb3_o0_fxp(ii+0)).int,...
          real(hb3_o0_fxp(ii+1)).int,imag(hb3_o0_fxp(ii+1)).int);
end
fclose(fid);

% Mixer0 output
fid = fopen('data/output_mixer_0.txt','w');
for ii = 1 : 2 : Nsamp*(3/2)*2*2*2
  fprintf(fid,'%d %d %d %d\n',...
          real(tuner_o0_fxp(ii+0)).int,imag(tuner_o0_fxp(ii+0)).int,...
          real(tuner_o0_fxp(ii+1)).int,imag(tuner_o0_fxp(ii+1)).int);
end
fclose(fid);

% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Save the ouput signals for sig 1
% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Resampler output

fid = fopen('data/output_resamp_1.txt','w');
% Two samples in one line
for ii = 1 : 2 :Nsamp*(3/2)
  fprintf(fid,'%d %d %d %d\n',...
          real(frc_o1_fxp(ii+0)).int,imag(frc_o1_fxp(ii+0)).int,...
          real(frc_o1_fxp(ii+1)).int,imag(frc_o1_fxp(ii+1)).int);
end
fclose(fid);

% One sample in one line
fid = fopen('data/output_resamp_1_one.txt','w');
for ii = 1 : Nsamp*(3/2)
  fprintf(fid,'%d %d \n',...
          real(frc_o1_fxp(ii+0)).int,imag(frc_o1_fxp(ii+0)).int);
end
fclose(fid);

% HB1 output
fid = fopen('data/output_hb1_1.txt','w');
for ii = 1 : 2 : Nsamp*(3/2)*2
  fprintf(fid,'%d %d %d %d\n',...
          real(hb1_o1_fxp(ii+0)).int,imag(hb1_o1_fxp(ii+0)).int,...
          real(hb1_o1_fxp(ii+1)).int,imag(hb1_o1_fxp(ii+1)).int);
end
fclose(fid);

% HB2 output
fid = fopen('data/output_hb2_1.txt','w');
for ii = 1 : 2 : Nsamp*(3/2)*2*2
  fprintf(fid,'%d %d %d %d\n',...
          real(hb2_o1_fxp(ii+0)).int,imag(hb2_o1_fxp(ii+0)).int,...
          real(hb2_o1_fxp(ii+1)).int,imag(hb2_o1_fxp(ii+1)).int);
end
fclose(fid);

% HB3 output
fid = fopen('data/output_hb3_1.txt','w');
for ii = 1 : 2 : Nsamp*(3/2)*2*2*2
  fprintf(fid,'%d %d %d %d\n',...
          real(hb3_o1_fxp(ii+0)).int,imag(hb3_o1_fxp(ii+0)).int,...
          real(hb3_o1_fxp(ii+1)).int,imag(hb3_o1_fxp(ii+1)).int);
end
fclose(fid);

% Mixer0 output
fid = fopen('data/output_mixer_1.txt','w');
for ii = 1 : 2 : Nsamp*(3/2)*2*2*2
  fprintf(fid,'%d %d %d %d\n',...
          real(tuner_o1_fxp(ii+0)).int,imag(tuner_o1_fxp(ii+0)).int,...
          real(tuner_o1_fxp(ii+1)).int,imag(tuner_o1_fxp(ii+1)).int);
end
fclose(fid);

% %%%
% Save the output
% %%%
fid = fopen('data/output_adder.txt','w');
for ii = 1 : 2 : Nsamp*(3/2)*2*2*2
  fprintf(fid,'%d %d %d %d\n',...
          real(sig_o_fxp(ii+0)).int,imag(sig_o_fxp(ii+0)).int,...
          real(sig_o_fxp(ii+1)).int,imag(sig_o_fxp(ii+1)).int);
end
fclose(fid);

% %%%

[pxx,f]   = pwelch(sig_i0/std(sig_i0)*sqrt(bw_sig0*fs),hanning(1e3),[],1e3,fs,'centered','power');
[pxx2,f2] = pwelch(sig_i1/std(sig_i1)*sqrt(bw_sig1*fs),hanning(1e3),[],1e3,fs,'centered','power');
[pxx3,f3] = pwelch(frc_o0/std(frc_o0)*sqrt(bw_sig0*fs),hanning(1e3),[],1e3,3/2*fs,'centered','power');
[pxx4,f4] = pwelch(frc_o1/std(frc_o1)*sqrt(bw_sig1*fs),hanning(1e3),[],1e3,3/2*fs,'centered','power');
[pxx5,f5] = pwelch(hb1_o0/std(hb1_o0)*sqrt(bw_sig0*fs),hanning(1e3),[],1e3,3/2*2*fs,'centered','power');
[pxx6,f6] = pwelch(hb1_o1/std(hb1_o1)*sqrt(bw_sig1*fs),hanning(1e3),[],1e3,3/2*2*fs,'centered','power');
[pxx7,f7] = pwelch(sig_o/std(sig_o)*sqrt((bw_sig0+bw_sig1)*fs),hanning(1e3),[],1e3,3/2*2*2*2*fs,'centered','power');
figure(5);
subplot(4,1,1);plot(f,10*log10(pxx));hold on;plot(f2,10*log10(pxx2));hold off;grid on;xlabel('Frequency (Hz)');ylabel('Magnitude (dB)');title('Input signals');legend('signal0','signal1');
subplot(4,1,2);plot(f3,10*log10(pxx3));hold on;plot(f4,10*log10(pxx4));hold off;grid on;xlabel('Frequency (Hz)');ylabel('Magnitude (dB)');title('FRC 3/2 outputs');legend('signal0','signal1');
subplot(4,1,3);plot(f5,10*log10(pxx5));hold on;plot(f6,10*log10(pxx6));hold off;grid on;xlabel('Frequency (Hz)');ylabel('Magnitude (dB)');title('HBF1 outputs');legend('signal0','signal1');
subplot(4,1,4);plot(f7,10*log10(pxx7));grid on;xlabel('Frequency (Hz)');ylabel('Magnitude (dB)');title('Sum Tuner outputs');
